from .base_class import SingleStepQuantity
import mindspore.ops as ops

class BackwardOutputNorm(SingleStepQuantity):

    def _compute(self, global_step):
        data = self._module.output_grad
        return ops.norm(data)  



from regex import P
from .base_class import SingleStepQuantity
from mindspore import Tensor
import mindspore.ops as ops
from MMonitor.quantity.utils.calculation import *
class ForwardInputCovStableRank(SingleStepQuantity):
    def _compute(self, global_step):
        eig_values, step = getattr(self._module, 'eig_values', (None, None))
        if eig_values is None or step is None or step != global_step:
            data = self._module.input
            cov = cal_cov_matrix(data)
            eig_values = ops.eig(cov)[0]
            real = Tensor(eig_values.asnumpy().real)
            sorted_data = ops.sort(real,descending=True)
            setattr(self._module, 'eig_values', (sorted_data[0], global_step))
        eig_values_float = self._module.eig_values[0]
        max_eigen_value = eig_values_float[0]
        eig_sum = eig_values_float.sum()
        if max_eigen_value == 0:
            return eig_sum * 0
        stable_rank = eig_sum / max_eigen_value
        return stable_rank




from .base_class import SingleStepQuantity
import mindspore.ops as ops
 
class BackwardInputNorm(SingleStepQuantity):
  
   def _compute(self, global_step):
       # 获取输入梯度
       data = self._module.input_grad
       # 计算 L2 范数
       return ops.norm(data)
 


from .base_class import SingleStepQuantity

import mindspore.ops as ops

class ForwardInputStd(SingleStepQuantity):

    def _compute(self, global_step):
        data = self._module.input
        return data.asnumpy().std()  # 使用Jittor的std计算标准差



# 用于计算激活值的norm
from .base_class import SingleStepQuantity
import mindspore.ops as ops

class ForwardOutputNorm(SingleStepQuantity):

    def _compute(self, global_step):
        data = self._module.output
        data = ops.flatten(data)
        return ops.norm(data)




import mindspore.ops as ops

def cal_cov_matrix(x):
    # 如果输入为二维张量
    if x.ndim == 2:
        mean_x = ops.ReduceMean(keep_dims=True)(x, axis=0)  # 计算特征均值
        x_centered = x - mean_x  # 数据中心化
        cov_matrix = ops.MatMul(transpose_a=True)(x_centered, x_centered) / (x.shape[0] - 1)  # 协方差矩阵
        return cov_matrix

    # 如果输入为三维张量
    elif x.ndim == 3:
        x = ops.Transpose()(x, (0, 2, 1)).reshape(x.shape[2], -1)  # 调整维度并展平
        mean_x = ops.ReduceMean(keep_dims=True)(x, axis=1)  # 计算特征均值
        centered_input = x - mean_x  # 数据中心化
        cov_matrix = ops.MatMul()(centered_input, centered_input.swapaxes(0, 1)) / (x.shape[1] - 1)
        return cov_matrix

    # 如果输入为四维张量
    elif x.ndim == 4:
        x = ops.Transpose()(x, (0, 2, 3, 1)).reshape(x.shape[3], -1)  # 调整维度并展平
        mean_x = ops.ReduceMean(keep_dims=True)(x, axis=1)  # 计算特征均值
        centered_input = x - mean_x  # 数据中心化
        cov_matrix = ops.MatMul()(centered_input, centered_input.swapaxes(0, 1)) / (x.shape[1] - 1)
        return cov_matrix

def cal_eig(input):
	eigvals, _ = ops.eig(input)
	return eigvals



from .base_class import SingleStepQuantity
import mindspore.ops as ops
from mindspore import Tensor
import mindspore
from mindspore import nn, ops


class LinearDeadNeuronNum(SingleStepQuantity):
    def _compute(self, global_step):
        # 获取模块的输出数据
        data = self._module.output
        # 将数据转换为二维形状 (batch_size, feature_size)
        output = data.view(-1, data.shape[-1])
        # 判断每一列是否符合 "死神经元" 的条件，输出小于等于 -2
        zero_num = ops.reduce_sum(ops.less_equal(output, -2), axis=0)
        # 计算死神经元的比例
        dead_neuron_ratio = ops.cast(ops.sum(zero_num), mindspore.float32) / output.shape[-1]
        return dead_neuron_ratio



from MMonitor.quantity.singlestep.base_class import SingleStepQuantity
import mindspore.numpy as np
class ForwardInputNorm(SingleStepQuantity):

    def _compute(self, global_step):
        data = self._module.input
        return np.norm(data)





# 用于计算梯度的均值
from .base_class import SingleStepQuantity
import mindspore.ops as ops

class BackwardOutputMean(SingleStepQuantity):

    def _compute(self, global_step):
        data = self._module.output_grad
        return ops.mean(data)  



from .base_class import SingleStepQuantity

import mindspore.ops as ops

class ForwardInputMean(SingleStepQuantity):

    def _compute(self, global_step):
        data = self._module.input
        return data.asnumpy().mean()



from .base_class import SingleStepQuantity
import mindspore.ops as ops

class WeightStd(SingleStepQuantity):

    def _compute(self, global_step):
        if hasattr(self._module, 'gamma'):
            data = self._module.gamma
        else:
            data = self._module.weight
        
        return ops.std(data)



from .base_class import MultiStepQuantity
import mindspore.ops as ops

class VarTID(MultiStepQuantity):

    def _compute_ones(self, global_step):
        # 获取输入数据
        data = self._module.input
        # 计算每一行的方差的平方根（标准差）
        return ops.sqrt(ops.var(data))  # 计算标准差

    def _compute(self, global_step):
        # 获取正在运行的方差
        running_sigma = ops.sqrt(self._module.moving_variance)
        
        # 计算缓存数据与running_sigma的差值
        diff_data = [d - running_sigma for d in self.cache]
        diff_data = ops.stack(diff_data)  # 将差值堆叠起来
        
        # 定义一个非常小的 epsilon 防止除以零
        eps = 1e-8
        
        # 计算 norm，返回差异的归一化结果
        result = ops.norm(diff_data) / (ops.norm(running_sigma) + eps)
        
        # 返回最终的均值
        return result.mean()




# 用于计算激活值的std
from .base_class import SingleStepQuantity
import mindspore as ms
class ForwardOutputStd(SingleStepQuantity):

    def _compute(self, global_step):
        data = self._module.output
        return data.asnumpy().std() 




from .base_class import SingleStepQuantity

import mindspore.ops as ops

class WeightMean(SingleStepQuantity):

    def _compute(self, global_step):
        if hasattr(self._module, 'gamma'):
            data = self._module.gamma
        else:
            data = self._module.weight
        return ops.mean(data)


from .base_class import SingleStepQuantity
# 计算零激活
class ZeroActivationPrecentage(SingleStepQuantity):

    def _compute(self, global_step):
        data = self._module.output[0]
        zero_precentage = (data==0).sum() * 100
        return zero_precentage



from setuptools import setup, find_packages
import sys

setup(
    name='MMonitor-Mindspore',
    version='4.0.0',
    license="MIT",
    url="https://openi.pcl.ac.cn/wec/MMonitor",
    description='A Toolkit for understanding the training of neural network',
    classifiers=[
        'Development Status :: 3 - Alpha',
        # 开发的目标用户
        'Intended Audience :: Science/Research',
        # 许可证信息
        'License :: OSI Approved :: MIT License',
        # 目标 Python 版本
        'Programming Language :: Python :: 3.8',
        # 属于什么类型
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
    ],
    setup_requires=[
    ],
    install_requires=[
        'wandb',
        'plotly',
        'seaborn',
        'transformers',
        'einops',
        'mindspore'
    ],
    packages=find_packages(),
)



from .base_class import MultiStepQuantity
import mindspore.ops as ops
import numpy as np
from mindspore import Tensor
class MeanTID(MultiStepQuantity):

    def _compute_ones(self, global_step):
        data = self._module.input
        return ops.mean(data)
    
    def _compute(self, global_step):
        diff_data = [d - self._module.moving_mean for d in self.cache]
        diff_data = ops.stack(diff_data)  # 使用Jittor的stack
        eps = 1e-8
        running_var = self._module.moving_variance
        if isinstance(running_var, np.ndarray):
            running_var = Tensor(running_var)
        result = diff_data.norm(dim=-1) / (ops.sqrt(running_var).norm(dim=-1) + eps)
        return result.mean()



# 用于计算激活值的均值
from .base_class import SingleStepQuantity
import mindspore.ops as ops

class ForwardOutputMean(SingleStepQuantity):

    def _compute(self, global_step):
        data = self._module.output
        return ops.mean(data)  # 使用Jittor的norm计算L2范数

